Set-StrictMode -Version 2.0

Function GetQuotedValue {
    param(
        [Parameter(Mandatory=$true)]
        [AllowNull()]
        [AllowEmptyString()]
        [string]$Value,
        [char[]]$TestValues = @(" ")
    )

    # break early if already quoted
    if($Value -match "\'.*\'|`".*\`""){
        return $Value
    }

    if([System.Linq.Enumerable]::Any($Value, [Func[char,bool]]{
        param($x) ($TestValues -contains $x) -or [char]::IsWhiteSpace($x) }))
    {
        return "`"$Value`""
    }

    $Value
}
function SafeTrim {
        param(
            [Parameter(Mandatory=$true, ValueFromPipeline=$true)]
            $value
        )
        if($value -is [string]){
            return $value.Trim()
        } else {
            return $value
        }
    }
Function CleanJoin {
    param(
        [string[]]$Values = @(),
        [string]$Separator = '.'
    )

    ($values | Where-Object { $_ }) -join $Separator
}

Function GetFileDate {
    Get-Date -f yyMMdd-hhmmssfffffff
}

Function SetPropertiesFromHash {
    [CmdletBinding(SupportsShouldProcess=$true)]
    Param(
        [Parameter(Mandatory=$true)]
        [string]$On,
        [Parameter(Mandatory=$true)]
        [System.Collections.IDictionary]$Properties,
        [string]$Path = $null
    )

    function FlattenHash {
        [CmdletBinding(SupportsShouldProcess=$true)]
        [OutputType([hashtable])]
        param(
            [Parameter(Mandatory=$true)]
            [ValidateScript({ Test-Path $_ -IsValid })]
            [string]$Context,
            [System.Collections.IDictionary]$Values = @{},
            $Prefix = ''
        )

        $results = @{}

        foreach($key in $Values.Keys)
        {
            if($PSCmdlet.ShouldProcess($Context, "Test for property $key")){
                $propArgs = @{ Path = $Context }
                if($prefix) { $propArgs.Name = $prefix }
                $found = Get-ItemProperty @propArgs | Get-Member -Name $key
                $nextKey = $found | ForEach-Object Name -ErrorAction SilentlyContinue
                if(-not $nextKey) {
                    $nextKey = $found | ForEach-Object ElementTagNam -ErrorAction SilentlyContinue
                }
                if(-not $nextKey) {
                    throw "Could not resolve property '$Prefix.$key' for item $On"
                }
            } else {
                Write-Warning "Cannot confirm property '$key' exists for non-existing path '$On'"
                $nextKey = $key
            }

            $nextPrefix = $nextKey
            if($Prefix) { $nextPrefix = "$Prefix.$nextPrefix" }
            if($Values.$key -is [System.Collections.IDictionary]) {
                $chain = FlattenHash -Values $Values.$key -Context $Context -Prefix $nextPrefix
                foreach($entry in $chain.keys) {
                    $results[$entry] = $chain.$entry
                }
            } else {
                $results[$nextPrefix] = $Values.$key
            }
        }

        return $results
    }

    try {
        $flattened = FlattenHash -Values $Properties -Context $On

        foreach($key in $flattened.Keys){
            $value = $flattened.$key
            if($PSCmdlet.ShouldProcess($On, "Set property $key => $value")) {
                # If the key contains 'password' do not write value
                if($key -notlike '*password*') {
                    WriteTaskInfo -MessageData "$key => $value" -Tag 'Setting'
                }
                Set-ItemProperty -Path $On -Name $key -Value $value
            }
        }
    } catch {
        Write-Error $_
    }
}

function IsAdmin {
    param(
        [Security.Principal.WindowsPrincipal]$User = (GetUser)
    )

    return HasBuiltInRole -User $User -Role 'Administrator'
}

function HasBuiltInRole {
    param(
        [Parameter(Mandatory=$true)]
        [Security.Principal.WindowsPrincipal]$User,
        [Parameter(Mandatory=$true)]
        [Security.Principal.WindowsBuiltInRole]$Role
    )
    return $User.IsInRole($Role)
}

function GetUser {
    param(
        [string]$UserId
    )

    if($UserId){
        try {
            return [Security.Principal.WindowsIdentity]::new($UserId)
        } catch {
            throw "Could not resolve user '$UserId'"
        }
    }

    return [Security.Principal.WindowsIdentity]::GetCurrent()
}

function ResolveLogPath {
    param(
        [Parameter(Mandatory=$true)]
        [string]$SourceFile,
        [Parameter(Mandatory=$true)]
        [ValidateScript({ Test-Path $_ -PathType Container -IsValid})]
        [string]$DestinationFolder
    )

    $basename = [IO.Path]::GetFileNameWithoutExtension($SourceFile)
    $filename = "$basename.{0:yyMMdd}" -f (Get-Date)

    $counter = 2
    $fullPath = Join-Path $DestinationFolder "$filename.log"
    while(Test-Path $fullPath){
        $fullPath = Join-Path $DestinationFolder "$filename ($($counter)).log"
        $counter++
    }
    return $fullpath
}

Function ShowNestedHelp {
    WriteInfo -MessageData "This is a nested prompt in the current session to allow diagnostics and correction of the above error." -InformationAction:Continue
    WriteInfo -MessageData "When you are ready type `"Exit`" to return to the Sitecore Install Framework and the check(s) will be performed again." -InformationAction:Continue
    WriteInfo -MessageData "Alternatively enter `"SkipRequire`" to bypass prerequisite." -InformationAction:Continue
}

Function SkipRequire {
    Exit -100
}

Function WriteLines{
    Param(
        [Parameter(Mandatory)]
        [ValidateScript({Test-Path $_ -IsValid})]
        [ValidateScript({[System.IO.Path]::IsPathRooted($_)})]
        [String]$File,
        [String[]]$Content,
        [System.Text.Encoding]$Encoding = [System.Text.Encoding]::UTF8,
        [Int32]$Retries = 10
    )

    $enc = $Encoding
    $crlf = $enc.GetBytes("`n")
    $tries = 0
    $fileLock = $false

    If (!(Test-Path -Path $File)){
        New-Item -Path $File
    }

    Do {
        Try {
            $fileLock = [System.IO.File]::Open($File, 'Open', 'ReadWrite', 'None')
        }
        Catch{
            Write-Warning -Message "Failed to get lock on file. $File"
            $tries++
            Start-Sleep -Milliseconds 100
        }
    } Until ($fileLock -or ($tries -eq $Retries))

    If ($tries -eq $Retries){
        Throw "Unable to get lock on file $File after $Retries attempt(s)."
    }

    $fileLock.SetLength(0)

    Foreach($line in $Content){
        $newLine = $enc.GetBytes($line)
        $fileLock.Write($newLine,0,$newLine.Length)
        $fileLock.Write($crlf,0,$crlf.Length)
    }

    $fileLock.Close()
}
# SIG # Begin signature block
# MIIXwQYJKoZIhvcNAQcCoIIXsjCCF64CAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUX0zHfUVXG7jqQLkSPtMikCK6
# krugghL8MIID7jCCA1egAwIBAgIQfpPr+3zGTlnqS5p31Ab8OzANBgkqhkiG9w0B
# AQUFADCBizELMAkGA1UEBhMCWkExFTATBgNVBAgTDFdlc3Rlcm4gQ2FwZTEUMBIG
# A1UEBxMLRHVyYmFudmlsbGUxDzANBgNVBAoTBlRoYXd0ZTEdMBsGA1UECxMUVGhh
# d3RlIENlcnRpZmljYXRpb24xHzAdBgNVBAMTFlRoYXd0ZSBUaW1lc3RhbXBpbmcg
# Q0EwHhcNMTIxMjIxMDAwMDAwWhcNMjAxMjMwMjM1OTU5WjBeMQswCQYDVQQGEwJV
# UzEdMBsGA1UEChMUU3ltYW50ZWMgQ29ycG9yYXRpb24xMDAuBgNVBAMTJ1N5bWFu
# dGVjIFRpbWUgU3RhbXBpbmcgU2VydmljZXMgQ0EgLSBHMjCCASIwDQYJKoZIhvcN
# AQEBBQADggEPADCCAQoCggEBALGss0lUS5ccEgrYJXmRIlcqb9y4JsRDc2vCvy5Q
# WvsUwnaOQwElQ7Sh4kX06Ld7w3TMIte0lAAC903tv7S3RCRrzV9FO9FEzkMScxeC
# i2m0K8uZHqxyGyZNcR+xMd37UWECU6aq9UksBXhFpS+JzueZ5/6M4lc/PcaS3Er4
# ezPkeQr78HWIQZz/xQNRmarXbJ+TaYdlKYOFwmAUxMjJOxTawIHwHw103pIiq8r3
# +3R8J+b3Sht/p8OeLa6K6qbmqicWfWH3mHERvOJQoUvlXfrlDqcsn6plINPYlujI
# fKVOSET/GeJEB5IL12iEgF1qeGRFzWBGflTBE3zFefHJwXECAwEAAaOB+jCB9zAd
# BgNVHQ4EFgQUX5r1blzMzHSa1N197z/b7EyALt0wMgYIKwYBBQUHAQEEJjAkMCIG
# CCsGAQUFBzABhhZodHRwOi8vb2NzcC50aGF3dGUuY29tMBIGA1UdEwEB/wQIMAYB
# Af8CAQAwPwYDVR0fBDgwNjA0oDKgMIYuaHR0cDovL2NybC50aGF3dGUuY29tL1Ro
# YXd0ZVRpbWVzdGFtcGluZ0NBLmNybDATBgNVHSUEDDAKBggrBgEFBQcDCDAOBgNV
# HQ8BAf8EBAMCAQYwKAYDVR0RBCEwH6QdMBsxGTAXBgNVBAMTEFRpbWVTdGFtcC0y
# MDQ4LTEwDQYJKoZIhvcNAQEFBQADgYEAAwmbj3nvf1kwqu9otfrjCR27T4IGXTdf
# plKfFo3qHJIJRG71betYfDDo+WmNI3MLEm9Hqa45EfgqsZuwGsOO61mWAK3ODE2y
# 0DGmCFwqevzieh1XTKhlGOl5QGIllm7HxzdqgyEIjkHq3dlXPx13SYcqFgZepjhq
# IhKjURmDfrYwggSjMIIDi6ADAgECAhAOz/Q4yP6/NW4E2GqYGxpQMA0GCSqGSIb3
# DQEBBQUAMF4xCzAJBgNVBAYTAlVTMR0wGwYDVQQKExRTeW1hbnRlYyBDb3Jwb3Jh
# dGlvbjEwMC4GA1UEAxMnU3ltYW50ZWMgVGltZSBTdGFtcGluZyBTZXJ2aWNlcyBD
# QSAtIEcyMB4XDTEyMTAxODAwMDAwMFoXDTIwMTIyOTIzNTk1OVowYjELMAkGA1UE
# BhMCVVMxHTAbBgNVBAoTFFN5bWFudGVjIENvcnBvcmF0aW9uMTQwMgYDVQQDEytT
# eW1hbnRlYyBUaW1lIFN0YW1waW5nIFNlcnZpY2VzIFNpZ25lciAtIEc0MIIBIjAN
# BgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAomMLOUS4uyOnREm7Dv+h8GEKU5Ow
# mNutLA9KxW7/hjxTVQ8VzgQ/K/2plpbZvmF5C1vJTIZ25eBDSyKV7sIrQ8Gf2Gi0
# jkBP7oU4uRHFI/JkWPAVMm9OV6GuiKQC1yoezUvh3WPVF4kyW7BemVqonShQDhfu
# ltthO0VRHc8SVguSR/yrrvZmPUescHLnkudfzRC5xINklBm9JYDh6NIipdC6Anqh
# d5NbZcPuF3S8QYYq3AhMjJKMkS2ed0QfaNaodHfbDlsyi1aLM73ZY8hJnTrFxeoz
# C9Lxoxv0i77Zs1eLO94Ep3oisiSuLsdwxb5OgyYI+wu9qU+ZCOEQKHKqzQIDAQAB
# o4IBVzCCAVMwDAYDVR0TAQH/BAIwADAWBgNVHSUBAf8EDDAKBggrBgEFBQcDCDAO
# BgNVHQ8BAf8EBAMCB4AwcwYIKwYBBQUHAQEEZzBlMCoGCCsGAQUFBzABhh5odHRw
# Oi8vdHMtb2NzcC53cy5zeW1hbnRlYy5jb20wNwYIKwYBBQUHMAKGK2h0dHA6Ly90
# cy1haWEud3Muc3ltYW50ZWMuY29tL3Rzcy1jYS1nMi5jZXIwPAYDVR0fBDUwMzAx
# oC+gLYYraHR0cDovL3RzLWNybC53cy5zeW1hbnRlYy5jb20vdHNzLWNhLWcyLmNy
# bDAoBgNVHREEITAfpB0wGzEZMBcGA1UEAxMQVGltZVN0YW1wLTIwNDgtMjAdBgNV
# HQ4EFgQURsZpow5KFB7VTNpSYxc/Xja8DeYwHwYDVR0jBBgwFoAUX5r1blzMzHSa
# 1N197z/b7EyALt0wDQYJKoZIhvcNAQEFBQADggEBAHg7tJEqAEzwj2IwN3ijhCcH
# bxiy3iXcoNSUA6qGTiWfmkADHN3O43nLIWgG2rYytG2/9CwmYzPkSWRtDebDZw73
# BaQ1bHyJFsbpst+y6d0gxnEPzZV03LZc3r03H0N45ni1zSgEIKOq8UvEiCmRDoDR
# EfzdXHZuT14ORUZBbg2w6jiasTraCXEQ/Bx5tIB7rGn0/Zy2DBYr8X9bCT2bW+IW
# yhOBbQAuOA2oKY8s4bL0WqkBrxWcLC9JG9siu8P+eJRRw4axgohd8D20UaF5Mysu
# e7ncIAkTcetqGVvP6KUwVyyJST+5z3/Jvz4iaGNTmr1pdKzFHTx/kuDDvBzYBHUw
# ggUrMIIEE6ADAgECAhAHplztCw0v0TJNgwJhke9VMA0GCSqGSIb3DQEBCwUAMHIx
# CzAJBgNVBAYTAlVTMRUwEwYDVQQKEwxEaWdpQ2VydCBJbmMxGTAXBgNVBAsTEHd3
# dy5kaWdpY2VydC5jb20xMTAvBgNVBAMTKERpZ2lDZXJ0IFNIQTIgQXNzdXJlZCBJ
# RCBDb2RlIFNpZ25pbmcgQ0EwHhcNMTcwODIzMDAwMDAwWhcNMjAwOTMwMTIwMDAw
# WjBoMQswCQYDVQQGEwJVUzELMAkGA1UECBMCY2ExEjAQBgNVBAcTCVNhdXNhbGl0
# bzEbMBkGA1UEChMSU2l0ZWNvcmUgVVNBLCBJbmMuMRswGQYDVQQDExJTaXRlY29y
# ZSBVU0EsIEluYy4wggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQC7PZ/g
# huhrQ/p/0Cg7BRrYjw7ZMx8HNBamEm0El+sedPWYeAAFrjDSpECxYjvK8/NOS9dk
# tC35XL2TREMOJk746mZqia+g+NQDPEaDjNPG/iT0gWsOeCa9dUcIUtnBQ0hBKsuR
# bau3n7w1uIgr3zf29vc9NhCoz1m2uBNIuLBlkKguXwgPt4rzj66+18JV3xyLQJoS
# 3ZAA8k6FnZltNB+4HB0LKpPmF8PmAm5fhwGz6JFTKe+HCBRtuwOEERSd1EN7TGKi
# xczSX8FJMz84dcOfALxjTj6RUF5TNSQLD2pACgYWl8MM0lEtD/1eif7TKMHqaA+s
# m/yJrlKEtOr836BvAgMBAAGjggHFMIIBwTAfBgNVHSMEGDAWgBRaxLl7Kgqjpepx
# A8Bg+S32ZXUOWDAdBgNVHQ4EFgQULh60SWOBOnU9TSFq0c2sWmMdu7EwDgYDVR0P
# AQH/BAQDAgeAMBMGA1UdJQQMMAoGCCsGAQUFBwMDMHcGA1UdHwRwMG4wNaAzoDGG
# L2h0dHA6Ly9jcmwzLmRpZ2ljZXJ0LmNvbS9zaGEyLWFzc3VyZWQtY3MtZzEuY3Js
# MDWgM6Axhi9odHRwOi8vY3JsNC5kaWdpY2VydC5jb20vc2hhMi1hc3N1cmVkLWNz
# LWcxLmNybDBMBgNVHSAERTBDMDcGCWCGSAGG/WwDATAqMCgGCCsGAQUFBwIBFhxo
# dHRwczovL3d3dy5kaWdpY2VydC5jb20vQ1BTMAgGBmeBDAEEATCBhAYIKwYBBQUH
# AQEEeDB2MCQGCCsGAQUFBzABhhhodHRwOi8vb2NzcC5kaWdpY2VydC5jb20wTgYI
# KwYBBQUHMAKGQmh0dHA6Ly9jYWNlcnRzLmRpZ2ljZXJ0LmNvbS9EaWdpQ2VydFNI
# QTJBc3N1cmVkSURDb2RlU2lnbmluZ0NBLmNydDAMBgNVHRMBAf8EAjAAMA0GCSqG
# SIb3DQEBCwUAA4IBAQBozpJhBdsaz19E9faa/wtrnssUreKxZVkYQ+NViWeyImc5
# qEZcDPy3Qgf731kVPnYuwi5S0U+qyg5p1CNn/WsvnJsdw8aO0lseadu8PECuHj1Z
# 5w4mi5rGNq+QVYSBB2vBh5Ps5rXuifBFF8YnUyBc2KuWBOCq6MTRN1H2sU5LtOUc
# Qkacv8hyom8DHERbd3mIBkV8fmtAmvwFYOCsXdBHOSwQUvfs53GySrnIYiWT0y56
# mVYPwDj7h/PdWO5hIuZm6n5ohInLig1weiVDJ254r+2pfyyRT+02JVVxyHFMCLwC
# ASs4vgbiZzMDltmoTDHz9gULxu/CfBGM0waMDu3cMIIFMDCCBBigAwIBAgIQBAkY
# G1/Vu2Z1U0O1b5VQCDANBgkqhkiG9w0BAQsFADBlMQswCQYDVQQGEwJVUzEVMBMG
# A1UEChMMRGlnaUNlcnQgSW5jMRkwFwYDVQQLExB3d3cuZGlnaWNlcnQuY29tMSQw
# IgYDVQQDExtEaWdpQ2VydCBBc3N1cmVkIElEIFJvb3QgQ0EwHhcNMTMxMDIyMTIw
# MDAwWhcNMjgxMDIyMTIwMDAwWjByMQswCQYDVQQGEwJVUzEVMBMGA1UEChMMRGln
# aUNlcnQgSW5jMRkwFwYDVQQLExB3d3cuZGlnaWNlcnQuY29tMTEwLwYDVQQDEyhE
# aWdpQ2VydCBTSEEyIEFzc3VyZWQgSUQgQ29kZSBTaWduaW5nIENBMIIBIjANBgkq
# hkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA+NOzHH8OEa9ndwfTCzFJGc/Q+0WZsTrb
# RPV/5aid2zLXcep2nQUut4/6kkPApfmJ1DcZ17aq8JyGpdglrA55KDp+6dFn08b7
# KSfH03sjlOSRI5aQd4L5oYQjZhJUM1B0sSgmuyRpwsJS8hRniolF1C2ho+mILCCV
# rhxKhwjfDPXiTWAYvqrEsq5wMWYzcT6scKKrzn/pfMuSoeU7MRzP6vIK5Fe7SrXp
# dOYr/mzLfnQ5Ng2Q7+S1TqSp6moKq4TzrGdOtcT3jNEgJSPrCGQ+UpbB8g8S9MWO
# D8Gi6CxR93O8vYWxYoNzQYIH5DiLanMg0A9kczyen6Yzqf0Z3yWT0QIDAQABo4IB
# zTCCAckwEgYDVR0TAQH/BAgwBgEB/wIBADAOBgNVHQ8BAf8EBAMCAYYwEwYDVR0l
# BAwwCgYIKwYBBQUHAwMweQYIKwYBBQUHAQEEbTBrMCQGCCsGAQUFBzABhhhodHRw
# Oi8vb2NzcC5kaWdpY2VydC5jb20wQwYIKwYBBQUHMAKGN2h0dHA6Ly9jYWNlcnRz
# LmRpZ2ljZXJ0LmNvbS9EaWdpQ2VydEFzc3VyZWRJRFJvb3RDQS5jcnQwgYEGA1Ud
# HwR6MHgwOqA4oDaGNGh0dHA6Ly9jcmw0LmRpZ2ljZXJ0LmNvbS9EaWdpQ2VydEFz
# c3VyZWRJRFJvb3RDQS5jcmwwOqA4oDaGNGh0dHA6Ly9jcmwzLmRpZ2ljZXJ0LmNv
# bS9EaWdpQ2VydEFzc3VyZWRJRFJvb3RDQS5jcmwwTwYDVR0gBEgwRjA4BgpghkgB
# hv1sAAIEMCowKAYIKwYBBQUHAgEWHGh0dHBzOi8vd3d3LmRpZ2ljZXJ0LmNvbS9D
# UFMwCgYIYIZIAYb9bAMwHQYDVR0OBBYEFFrEuXsqCqOl6nEDwGD5LfZldQ5YMB8G
# A1UdIwQYMBaAFEXroq/0ksuCMS1Ri6enIZ3zbcgPMA0GCSqGSIb3DQEBCwUAA4IB
# AQA+7A1aJLPzItEVyCx8JSl2qB1dHC06GsTvMGHXfgtg/cM9D8Svi/3vKt8gVTew
# 4fbRknUPUbRupY5a4l4kgU4QpO4/cY5jDhNLrddfRHnzNhQGivecRk5c/5CxGwcO
# kRX7uq+1UcKNJK4kxscnKqEpKBo6cSgCPC6Ro8AlEeKcFEehemhor5unXCBc2XGx
# DI+7qPjFEmifz0DLQESlE/DmZAwlCEIysjaKJAL+L3J+HNdJRZboWR3p+nRka7Lr
# ZkPas7CM1ekN3fYBIM6ZMWM9CBoYs4GbT8aTEAb8B4H6i9r5gkn3Ym6hU/oSlBiF
# LpKR6mhsRDKyZqHnGKSaZFHvMYIELzCCBCsCAQEwgYYwcjELMAkGA1UEBhMCVVMx
# FTATBgNVBAoTDERpZ2lDZXJ0IEluYzEZMBcGA1UECxMQd3d3LmRpZ2ljZXJ0LmNv
# bTExMC8GA1UEAxMoRGlnaUNlcnQgU0hBMiBBc3N1cmVkIElEIENvZGUgU2lnbmlu
# ZyBDQQIQB6Zc7QsNL9EyTYMCYZHvVTAJBgUrDgMCGgUAoHAwEAYKKwYBBAGCNwIB
# DDECMAAwGQYJKoZIhvcNAQkDMQwGCisGAQQBgjcCAQQwHAYKKwYBBAGCNwIBCzEO
# MAwGCisGAQQBgjcCARUwIwYJKoZIhvcNAQkEMRYEFGpueff/V0LrEFwvlIHZUFGR
# 5wjJMA0GCSqGSIb3DQEBAQUABIIBAJsRBZI0knkNYmb3q67973PsISJ7kG91Rxa1
# TCcphRQKYqbytAA+beOG/LF23HfDNoz4ODaWFnndMFutPCplYBV2likG+x8kUji4
# H5CgRpRlxpcCCm+2RXHrAdTn6rkcFLC+HPxino8r8JXq0wtM00EnnyV/VRmsR3Gf
# 0fIEVb+NDzRoLPspLxTeSkrE5tiNruyWB/aamYzoKfnQu08DeQRvRumTb8g7OHP7
# liSJPzr5KNInkXJnIheWsqgu9+uptNry5V2LVyM0iQUG/8LMpMyWJnkN0wLyKjs5
# TfAfWRmIqc9Ssz7+zS50DsTwmELQIfnhcf3Gu/dSOfI/kwSWiByhggILMIICBwYJ
# KoZIhvcNAQkGMYIB+DCCAfQCAQEwcjBeMQswCQYDVQQGEwJVUzEdMBsGA1UEChMU
# U3ltYW50ZWMgQ29ycG9yYXRpb24xMDAuBgNVBAMTJ1N5bWFudGVjIFRpbWUgU3Rh
# bXBpbmcgU2VydmljZXMgQ0EgLSBHMgIQDs/0OMj+vzVuBNhqmBsaUDAJBgUrDgMC
# GgUAoF0wGAYJKoZIhvcNAQkDMQsGCSqGSIb3DQEHATAcBgkqhkiG9w0BCQUxDxcN
# MjAwNjI1MjEyNzE1WjAjBgkqhkiG9w0BCQQxFgQUC36AR7VuumKzHvODZymieZLL
# bCowDQYJKoZIhvcNAQEBBQAEggEAANd4PIFnP5R86BtqsB04JkhseEcUBuudfnAT
# zUiVw7HARCJyozTyBJQL5J+iD4iHKXkchUe3TWKoK6nNQBckQ6w2npVx/gXwwzEh
# Fp1uM0b1EW8NiuziHv/l1QcuN+ulTsC0nbJUSrTEPB3JPpfm2gRnA019zdH+YuWX
# rgFDOWmBP5Zhvw6YI0fsXEfZPjYgPNtU8PMNAQvHqjuuVyL2+OYilS8y7sUSlNLA
# WuonDE2xFBlbH9VN+z4pO2kgSYxok+nTPWmuD53wOjiv4zQWIDFm+OVlYqyMSAIZ
# WVcS6NbSlwniWCZsO1z0TGvbrOMroO1JEOpuHXSPtJL/YIlSvA==
# SIG # End signature block
